"""
    MIT License

    Copyright (c) 2022 kian Ahmadian

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from __future__ import annotations
import asyncio
from typing import Callable, Dict, Tuple, List, Optional
from builtins import enumerate, reversed
from .error import NotFound
from bale import (Message, Update, User, Components, RemoveComponents, Chat, Price, ChatMember, HTTPClient, Updater,
                  Photo, Document)


__all__ = (
    "Bot"
)


class _Loop:
    __slots__ = ()
    

_loop = _Loop()

class Bot:
    """This object represents a Bale Bot.

    Parameters
    ----------
        token: str 
            Bot Token
    """
    __slots__ = (
        "loop",
        "token",
        "loop",
        "events",
        "listeners",
        "updater",
        "_user",
        "http",
        "_closed"
    )

    def __init__(self, token: str):
        self.loop = _loop
        self.token = token
        self.http: HTTPClient = HTTPClient(loop=self.loop, token=token)
        self._user = None
        self.updater = Updater(self)
        self.events: Dict[str, List[Callable]] = {}
        self.listeners: Dict[str, List[Tuple[asyncio.Future, Callable[..., bool]]]] = {}
        self._closed = False

    def listen(self, event_name):
        """Register a Event"""
        return lambda func: self.add_event(event_name, func)

    def add_event(self, event: str, function):
        """Register an Event with event name"""
        if not asyncio.iscoroutinefunction(function):
            raise TypeError(f"{function.__name__} is not a Coroutine Function")

        if not self.events.get(event):
            self.events[event] = list()

        self.events[event].append(function)

    def remove_event(self, event: str, function = None):
        """Register an Event with event name"""
        result = self.events.get(event)
        if not result:
            raise TypeError(f"{event} not in Events")

        if not function:
            del self.events[event]
            return

        if not function in result:
            raise TypeError(f"{function.__name__} not in Event Functions")

        del self.events[event][function]

    def wait_for(self, event_name: str, *, check = None, timeout = None):
        """Wait for an event"""
        self.loop: asyncio.AbstractEventLoop
        future = self.loop.create_future()
        event_name = event_name.lower()
        if not check:
            check = lambda: True

        listeners = self.listeners.get(event_name)
        if not listeners:
            listeners = []
            self.listeners[event_name] = listeners

        listeners.append((future, check))
        return asyncio.wait_for(future, timeout = timeout)

    @property
    def user(self) -> "User" or None:
        """Represents the connected client. ``None`` if not logged in"""
        return self._user

    async def __aenter__(self):
        loop = asyncio.get_running_loop()
        self.loop = loop
        self.http.loop = loop
        await self.http.start()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        pass

    async def close(self):
        """Close http Events and bot"""
        await self.http.close()
        self._closed = True

    def is_closed(self):
        """:class:`bool`: Connection Status"""
        return self._closed

    async def run_event(self, core, event_name, *args, **kwargs):
        try:
            await core(*args, **kwargs)
        except Exception as ext:
            await self.on_error(event_name, ext)

    def call_to_run_event(self, core, event_name, *args, **kwargs):
        task = self.run_event(core, event_name, *args, **kwargs)
        self.loop: asyncio.AbstractEventLoop
        return self.loop.create_task(task, name = event_name)

    def dispatch(self, event_name, /, *args, **kwargs):
        method = "on_" + event_name
        listeners = self.listeners.get(event_name)
        if listeners:
            removed = []
            for index, (future, check) in enumerate(listeners):
                if future.cancelled():
                    removed.append(index)
                    continue
                try:
                    result = check(*args)
                except Exception as __exception:
                    future.set_exception(__exception)
                    removed.append(index)
                else:
                    if result:
                        if len(args) == 0:
                            future.set_result(None)
                        elif len(args) == 1:
                            future.set_result(args[0])
                        else:
                            future.set_result(args)
                        removed.append(index)

            if len(listeners) == len(removed):
                self.listeners.pop(event_name)
            else:
                for index in reversed(removed):
                    del listeners[index]

        events_core = self.events.get(method)
        if events_core:
            for event_core in events_core:
                self.call_to_run_event(event_core, method, *args, **kwargs)

    async def on_error(self, event_name, error):
        """a Event for get errors when exceptions"""
        print("error", event_name, error)


    async def get_bot(self) -> User:
        """Get bot information

        Returns
        -------
            :class:`Bale.User` :
                Bot User information.
        Raises
        ------
            :class:`Bale.Error`
        """
        response = await self.http.get_bot()
        return User.from_dict(data=response.result, bot=self)

    async def delete_webhook(self) -> bool:
        """This service is used to remove the webhook set for the bot.

        Returns
        -------
            bool:
                ``True`` else ``False`` if not done
        """
        response = await self.http.delete_webhook()
        return response.result or False

    async def send_message(self, chat: "Chat" | "User", text: str, *, components: Optional["Components" | "RemoveComponents"] =None, reply_to_message: Optional["Message"] = None) -> "Message":
        """This service is used to send text messages.

        Parameters
        ----------
            chat: :class:`bale.Chat` | :class:`bale.User`
                Chat
            text: :class:`str`
                Message Text
            components: Optional[:class:`bale.Components` | :class:`bale.RemoveComponents`]
                Message Components
            reply_to_message: Optional[:class:`bale.Message`]
                Reply to a Message

        Raises
        ------
            :class:`bale.Error`
        Returns
        -------
            :class:`bale.Message`
                The Message
        """
        if not isinstance(chat, (Chat, User)):
            raise TypeError(
                "chat param must be type of Chat or User"
            )

        if components:
            if not isinstance(components, (Components, RemoveComponents)):
                raise TypeError(
                    "components param must be type of Components or RemoveComponents"
                )
            components = components.to_dict()


        if reply_to_message:
            if not isinstance(reply_to_message, Message):
                raise TypeError(
                    "reply_to_message param must be type of Message"
                )
            reply_to_message = reply_to_message.message_id

        response = await self.http.send_message(str(chat.chat_id), text, components=components, reply_to_message_id=reply_to_message)
        return Message.from_dict(data=response.result, bot=self)

    async def send_document(self, chat: "Chat" | "User", document: bytes | str | "Document", *, caption: Optional[str] = None, reply_to_message: Optional["Message"] = None) -> "Message":
        """This service is used to send document.

        Parameters
        ----------
        chat: :class:`bale.Chat` | :class:`bale.User`
            Chat
        document: :class:`bytes` | :class:`str` | :class:`bale.Document`
            Document
        caption: Optional[:class:`str`]
            Message caption
        reply_to_message: Optional[:class:`bale.Message`]
            Reply to a Message
        Raises
        ------
            :class:`bale.Error`

        Returns
        --------
            :class:`bale.Message`
                The Message.
        """
        if not isinstance(chat, (Chat, User)):
            raise TypeError(
                "chat param must be type of Chat or User"
            )

        if not isinstance(document, (bytes, str, Document)):
            raise TypeError(
                "document param must be type of bytes, str or Document"
            )

        if reply_to_message:
            if not isinstance(reply_to_message, Message):
                raise TypeError(
                    "reply_to_message param must be type of Message"
                )
            reply_to_message = reply_to_message.message_id

        if isinstance(document, Document):
            document = document.file_id

        response = await self.http.send_document(chat.chat_id, document, caption=caption, reply_to_message_id=reply_to_message)
        return Message.from_dict(data=response.result, bot=self)

    async def send_photo(self, chat: "Chat" | "User", photo: bytes | str | "Photo", *, caption: Optional[str] = None, reply_to_message: Optional["Message"] = None) -> "Message":
        """This service is used to send photo.

        Parameters
        ----------
        chat: :class:`bale.Chat` | :class:`bale.User`
            Chat
        photo: :class:`bytes` | :class:`str` | :class:`bale.Photo`
            Photo
        caption: Optional[:class:`str`]
            Message caption
        reply_to_message: Optional[:class:`bale.Message`]
            Reply to a Message
        Raises
        ------
            :class:`bale.Error`

        Returns
        --------
            :class:`bale.Message`
                The Message.
        """
        if not isinstance(chat, (Chat, User)):
            raise TypeError(
                "chat param must be type of Chat or User"
            )

        if not isinstance(photo, (bytes, str, Photo)):
            raise TypeError(
                "photo param must be type of bytes, str or Photo"
            )

        if isinstance(photo, Photo):
            photo = photo.file_id

        if reply_to_message:
            if not isinstance(reply_to_message, Message):
                raise TypeError(
                    "reply_to_message param must be type of Message"
                )
            reply_to_message = reply_to_message.message_id

        if caption and not isinstance(caption, str):
            raise TypeError(
                "caption param must be type of str"
            )

        response = await self.http.send_photo(str(chat.chat_id), photo, caption=caption, reply_to_message_id=reply_to_message)
        return Message.from_dict(data=response.result, bot=self)

    async def send_invoice(self, chat: "Chat" | "User", title: str, description: str, provider_token: str, prices: List["Price"], *, photo_url: Optional[str] = None, need_name: Optional[bool] = False, need_phone_number: Optional[bool] = False, need_email: Optional[bool] = False, need_shipping_address: Optional[bool] = False, is_flexible: Optional[bool] = True) -> Message:
        """You can use this service to send money request messages.

        Parameters
        ----------
        chat: :class:`bale.Chat` | :class:`bale.User`
            Chat
        title: str
            Invoice Title
        description: str
            Invoice Description
        provider_token: str
            You can use 3 methods to receive money: 1.Card number 2. Port number and acceptor number 3. Wallet number "Bale"
        prices: List[:class:`bale.Price`]
            A list of prices.
        photo_url: str
            Photo URL of Invoice. Defaults to None.
        need_name: bool
            Get a name from "User"?. Defaults to False.
        need_phone_number: bool
            Get a Phone number from "User"?. Defaults to False.
        need_email: bool
            Get a Email from "User"?. Defaults to False.
        need_shipping_address: bool
            Get a Shipping Address from "User"?. Defaults to False.
        is_flexible: bool
            Is the Invoice Photo Flexible to the Payment button?. Defaults to True.

        Returns
        -------
            :class:`bale.Message`
        """
        if not isinstance(chat, (Chat, User)):
            raise TypeError(
                "chat param must be type of Chat or User"
            )

        if not isinstance(photo_url, str):
            raise TypeError(
                "photo_url param must be type of str"
            )

        if not isinstance(need_name, bool):
            raise TypeError(
                "need_name param must be type of boolean"
            )

        if not isinstance(need_phone_number, bool):
            raise TypeError(
                "need_phone_number param must be type of boolean"
            )

        if not isinstance(need_email, bool):
            raise TypeError(
                "need_email param must be type of boolean"
            )

        if not isinstance(need_shipping_address, bool):
            raise TypeError(
                "need_shipping_address param must be type of boolean"
            )

        if not isinstance(is_flexible, bool):
            raise TypeError(
                "is_flexible param must be type of boolean"
            )

        prices = [price.to_dict() for price in prices if isinstance(price, Price)]
        response = await self.http.send_invoice(str(chat.chat_id), title, description, provider_token, prices, photo_url, need_name, need_phone_number, need_email, need_shipping_address, is_flexible)
        return Message.from_dict(data=response.result, bot=self)

    async def edit_message(self, chat: "Chat" | "User", message: "Message", text: str, *, components: "Components" | "RemoveComponents"=None) -> "Message":
        """You can use this service to edit text messages that you have already sent through the arm.

        Parameters
        ----------
            chat: :class:`bale.Chat`
                chat
            message: :class:`bale.Message`
                message
            text: str
                new content for message.
            components: Optional[:class:`bale.Components` | :class:`bale.RemoveComponents`]
                message components
        Raises
        ------
            :class:`bale.Error`
        Returns
        -------
            :class:`dict`
        """
        if not isinstance(chat, (Chat, User)):
            raise TypeError(
                "chat param must be type of Chat or User"
            )

        if not isinstance(message, Message):
            raise TypeError(
                "message_id param must be type of Message or str"
            )

        if components and not isinstance(components, (Components, RemoveComponents)):
            raise TypeError(
                "components param must be type of Components or RemoveComponents"
            )

        if components:
            components = components.to_dict()

        response = await self.http.edit_message(chat.chat_id, message.message_id, text, components=components)
        return response.result

    async def delete_message(self, chat: "Chat" | "User", message: "Message") -> bool:
        """You can use this service to delete a message that you have already sent through the arm.

        .. warning::
            In Channel or Group:
                If it is a group or channel Manager, it can delete a message from (group or channel).

            In private message (PV):
                If the message was sent by a bot, it can be deleted with this method

        Parameters
        ----------
            chat: :class:`bale.Chat`
                chat
            message: :class:`bale.Message`
                message
        Returns
        -------
            bool:
                ``True`` if done else ``False``
        """
        if not isinstance(chat, (Chat, User)):
            raise TypeError(
                "chat param must be type of Chat or User"
            )

        if not isinstance(message, Message):
            raise TypeError(
                "message param must be type of str or Message"
            )
        response = await self.http.delete_message(str(chat.chat_id), message.message_id)
        return response.result or False

    async def get_chat(self, chat_id: int | str) -> Chat | None:
        """This service can be used to receive personal information that has previously interacted with the arm.

        Parameters
        ----------
            chat_id: int | str
                chat id
        Raises
        ------
            :class:`bale.Error`
        Returns
        -------
            Optional[:class:`bale.Chat`]
                The chat or ``None`` if not found.
        """
        if not isinstance(chat_id, (int, str)):
            raise TypeError(
                "chat_id param must be type of int or str"
            )

        try:
            response = await self.http.get_chat(str(chat_id))
        except NotFound:
            return None
        else:
            return Chat.from_dict(response.result, bot=self)

    async def get_user(self, user_id: int | str) -> "User" | None:
        """This Method almost like :class:`bale.Bot.get_chat` , but this a filter that only get user.

        Parameters
        ----------
            user_id: int
                user id
        Raises
        ------
            :class:`bale.Error`
        Returns
        -------
            Optional[:class:`bale.User`]
                The user or ``None`` if not found.
        """
        if not isinstance(user_id, (int, str)):
            raise TypeError(
                "user_id param must be type of int or str"
            )

        chat = await self.get_chat(user_id)
        if chat and chat.type.is_private_chat():
            return User.from_dict(chat.to_dict(), self)

        return None

    async def get_chat_member(self, chat: "Chat", user_id: int) -> "ChatMember" | None:
        """
        Parameters
        ----------
            chat: :class:`bale.Chat`
                chat
            user_id: int
                user

        Raises
        ------
            :class:`bale.Error`

        Returns
        -------
            Optional[:class:`bale.ChatMember`]:
                The chat member or ``None`` if not found.
        """
        if not isinstance(chat, Chat):
            raise TypeError(
                "chat param must be type of Chat"
            )

        if not isinstance(user_id, int):
            raise TypeError(
                "user_id param must be type of int"
            )

        try:
            response = await self.http.get_chat_member(chat_id=str(chat.chat_id), member_id=str(user_id))
        except NotFound:
            return None
        else:
            return ChatMember.from_dict(response.result)

    async def invite_to_chat(self, chat: "Chat", user: "User") -> bool:
        """Invite user to the chat

        Parameters
        ----------
            chat: :class:`bale.Chat`
                chat
            user: :class:`bale.User`
                user

        Raises
        ------
            :class:`bale.Error`
        
        Returns
        -------
            bool:
                ``True`` when user added to chat else ``False``
        """
        if not isinstance(chat, Chat):
            raise TypeError(
                "chat param must be type of Chat"
            )

        if not isinstance(user, User):
            raise TypeError(
                "user param must be type of User"
            )

        response = await self.http.invite_to_chat(str(chat.chat_id), str(user.user_id))
        return response.result or False

    async def leave_chat(self, chat: "Chat"):
        """Leave bot from a Chat

        Parameters
        ----------
            chat: :class:`bale.Chat`
                chat
        
        Raises
        ------
            :class:`bale.Error`

        Returns
        -------
            bool:
                ``True`` when bot leaved from chat else ``False``
        """
        if not isinstance(chat, Chat):
            raise TypeError(
                "chat param must be type of Chat"
            )
        response = await self.http.leave_chat(str(chat.chat_id))
        return response.result or False

    async def get_chat_members_count(self, chat: "Chat") -> int | None:
        """
        Parameters
        ----------
            chat: :class:`bale.Chat` 

        Raises
        ------
            :class:`bale.Error`
                Group ID

        Returns
        --------
            Optional[int]:
                int or ``None`` if chat not found.
        """
        if not isinstance(chat, Chat):
            raise TypeError(
                "chat param must be type of Chat"
            )

        response = await self.http.get_chat_members_count(str(chat.chat_id))
        return response.result

    async def get_chat_administrators(self, chat: "Chat") -> list["ChatMember"] | None:
        """This service can be used to display admins of a group or channel.

        Parameters
        ----------
            chat: :class:`bale.Chat` 
                Group id
        Raises
        ------
            :class:`bale.Error`
        Returns
        -------
            Optional[List[:class:`bale.ChatMember`]]:
                list of chat member or ``None`` if chat not found.
        """
        if not isinstance(chat, Chat):
            raise TypeError(
                "chat param must be type of Chat"
            )

        response = await self.http.get_chat_administrators(chat.chat_id)
        return [ChatMember.from_dict(data=member_payload) for member_payload in response.result or list()]

    async def get_updates(self, offset: int = None, limit: int = None) -> list["Update"] | None:
        if offset and not isinstance(offset, int):
            raise TypeError(
                "offset param must be int"
            )

        if limit and not isinstance(limit, int):
            raise TypeError(
                "limit param must be int"
            )

        response = await self.http.get_updates(offset, limit)
        return [Update.from_dict(data=update_payload, bot=self) for update_payload in response.result or list() if not offset or (offset and update_payload.get("update_id") > offset)]

    async def connect(self, sleep_after_every_get_updates):
        self._user = await self.get_bot()
        await self.updater.start(sleep_after_every_get_updates = sleep_after_every_get_updates)

    def run(self, sleep_after_every_get_updates = None):
        """Run bot and https"""
        async def main():
            async with self:
                await self.connect(sleep_after_every_get_updates = sleep_after_every_get_updates)

        asyncio.run(main())



from __future__ import annotations


class Price:
	"""This object shows a Price

    Attributes
    ----------
        label: Optional[:class:`str`]
            Label Price.
        amount: Optional[:class:`int`]
            Amount Price.
    """
	__slots__ = ("label", "amount")

	def __init__(self, label: str = None, amount: int = None):
		self.label = label
		self.amount = amount

	@classmethod
	def from_dict(cls, data):
		return cls(label=data["label"], amount=data["amount"])

	def to_dict(self):
		data: dict[str, int | str] = {
			"label": self.label,
			"amount": self.amount
		}
		return data



from bale import (AdminPermissions, User)


class MemberRole:
    """Member Role"""
    OWNER = "creator"
    ADMIN = "administrator"
    __slots__ = ("_role",)

    def __init__(self, _role: str):
        self._role = _role

    def is_owner(self):
        """Member have owner role?"""
        return self._role == self.OWNER

    def is_admin(self):
        """Member have admin role?"""
        return self._role == self.ADMIN


class ChatMember:
    """This object shows a member in chat

    Attributes
    ----------
        role: :class:`bale.MemberRole`
            User Role
        user: :class:`bale.User`
            User
        permissions: :class:`bale.AdminPermissions`
            User Permissions
    """
    __slots__ = (
        "role", "_user", "permissions"
    )

    def __init__(self, role: "MemberRole" = None, user=None, permissions=None):
        self.role = role
        self._user = user
        self.permissions = permissions


    @property
    def user(self):
        return self._user

    @classmethod
    def from_dict(cls, data: dict):
        return cls(permissions=AdminPermissions.from_dict(data), user=User.from_dict(data.get("user")), role=MemberRole(data.get("status")))



"""
    MIT License

    Copyright (c) 2022 kian Ahmadian

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""
from bale.version import BALE_API_BASE_URL, BALE_API_FILE_URL
import asyncio
import aiohttp
from typing import Any
from ..error import (NetworkError, TimeOut, NotFound, Forbidden, APIError, BaleError, HTTPClientError, RateLimited)
from .parser import ResponseParser
from .utils import ResponseStatusCode
from collections import deque

__all__ = ("HTTPClient", "Route")


class RateLimit:
	__slots__ = (
		"_loop",
		"_requests",
		"_sleeping",
		"_has_rate_limit"
	)
	def __init__(self, loop):
		self._loop = loop
		self._requests: deque[asyncio.Future[Any]] = deque()
		self._has_rate_limit = False

	def enable(self):
		self._has_rate_limit = True

	def _wake_next(self):
		while self._requests:
			future = self._requests.popleft()

			if not future.done():
				future.set_result(None)
				break

	def _wake(self):
		while self._requests:
			future = self._requests.popleft()
			if not future.done():
				future.set_result(None)

	async def new_requests(self):
		while self._has_rate_limit:
			future = self._loop.create_future()
			self._requests.append(future)

			try:
				await future
			except:
				future.cancel()
				if not self._has_rate_limit:
					self._wake_next()

	async def __aenter__(self):
		await self.new_requests()
		return self

	async def __aexit__(self, exc_type, exc_val, exc_tb):
		has_error = (True if exc_type or exc_val or exc_type else False)

		if (self._has_rate_limit and not has_error) or (self._has_rate_limit and has_error and len(self._requests) <= 0):
			self._has_rate_limit = False

		if len(self._requests) > 0:
			self._wake_next()

	def __bool__(self):
		return self._has_rate_limit

class Route:
	"""Route Class for http"""
	BASE = BALE_API_BASE_URL
	BASE_FILE = BALE_API_FILE_URL

	__slots__ = (
		"method",
		"path",
		"token",
		"_base"
	)

	def __init__(self, method: str, path: str, token: str):
		if not isinstance(token, str):
			raise TypeError("token param must be str.")
		self.method = method
		self.path = path
		self.token = token
		self._base = self.BASE

	@property
	def url(self):
		"""Export url"""
		return self.BASE + "bot" + self.token + "/" + self.path

	def set_base(self, _value):
		"""Set base url for route"""
		self._base = _value

class HTTPClient:
	"""Send a Request to BALE API Server"""

	__slots__ = (
		"_loop",
		"token",
		"__session",
		"rate_limit"
	)

	def __init__(self, loop, token=None):
		self.__session = None
		self._loop: asyncio.AbstractEventLoop = loop
		self.token = token
		self.rate_limit = RateLimit(self.loop)

	def is_closed(self):
		return self.__session is None

	@property
	def loop(self):
		return self._loop

	@loop.setter
	def loop(self, _value):
		self._loop = _value
		self.rate_limit._loop = _value

	def reload_session(self):
		"""Reset Session"""
		if self.__session and self.__session.closed:
			self.__session = aiohttp.ClientSession(loop=self.loop, connector=aiohttp.TCPConnector(limit=5, keepalive_timeout=20.0))

	async def start(self):
		"""Start Http client"""
		self.__session = aiohttp.ClientSession(loop=self.loop, connector=aiohttp.TCPConnector(limit=5, keepalive_timeout=20.0))

	async def close(self):
		"""Close Session connection"""
		if self.__session:
			await self.__session.close()
			self.__session = None

	async def request(self, route: Route, **kwargs):
		url = route.url
		method = route.method
		async with self.rate_limit:
			for tries in range(5):
				try:
					async with self.__session.request(method=method, url=url, **kwargs) as response:
						response: aiohttp.ClientResponse = response
						parsed_response = await ResponseParser.from_response(response)
						if response.status == ResponseStatusCode.NOT_FOUND:
							raise NotFound()
						elif not parsed_response.ok or response.status == ResponseStatusCode.NOT_INCORRECT:
							if parsed_response.description == HTTPClientError.USER_OR_CHAT_NOT_FOUND:
								raise NotFound("User or Chat not Found")
							elif parsed_response.description == HTTPClientError.RATE_LIMIT or parsed_response.description == HTTPClientError.LOCAL_RATE_LIMIT:
								if tries >= 4:
									raise RateLimited()

								if bool(self.rate_limit):
									await self.rate_limit.new_requests()
								else:
									self.rate_limit.enable()
									await asyncio.sleep(tries + 1)
								continue
							elif parsed_response.description == HTTPClientError.PERMISSION_DENIED:
								raise Forbidden()

							raise APIError(
									str(parsed_response.error_code), parsed_response.description
								)

						elif response.status == ResponseStatusCode.OK:
							return parsed_response
				except aiohttp.client_exceptions.ClientConnectorError as error:
					raise NetworkError(str(error))
				except aiohttp.client_exceptions.ServerTimeoutError:
					raise TimeOut()
				except aiohttp.client_exceptions.ClientOSError as error:
					raise BaleError(error)

	async def get_file(self, file_id):
		async with self.__session.get(BALE_API_FILE_URL + "/" + "bot" + self.token + "/" + file_id) as response:
			if response.status == ResponseStatusCode.OK:
				return await response.read()
			elif response.status == ResponseStatusCode.NOT_FOUND:
				raise NotFound("Document is not Found")
			elif response.status == ResponseStatusCode.PERMISSION_DENIED:
				raise Forbidden()
			else:
				error_payload = await response.json()
				raise APIError(0, "UNKNOWN ERROR: {}".format(error_payload))

		raise RuntimeError("failed to get file")

	def send_message(self, chat_id, text, *, components=None, reply_to_message_id=None):
		payload = {
			"chat_id": chat_id,
			"text": text
		}
		if components:
			payload["reply_markup"] = components
		if reply_to_message_id:
			payload["reply_to_message_id"] = reply_to_message_id

		return self.request(Route("POST", "sendMessage", self.token), json=payload)

	def send_document(self, chat_id, document, *, caption=None, reply_to_message_id=None):
		payload = {
			"chat_id": chat_id,
			"document": document
		}
		if caption:
			payload["caption"] = caption

		if reply_to_message_id:
			payload["reply_to_message_id"] = reply_to_message_id

		return self.request(Route("POST", "Senddocument", self.token), data=payload)

	def send_photo(self, chat_id, photo, *, caption=None, reply_to_message_id=None):
		payload = {
			"chat_id": chat_id,
			"photo": photo
		}
		if caption:
			payload["caption"] = caption
		if reply_to_message_id:
			payload["reply_to_message_id"] = reply_to_message_id

		return self.request(Route("POST", "SendPhoto", self.token), data=payload)

	def send_invoice(self, chat_id, title, description, provider_token, prices, photo_url=None, need_name=False, need_phone_number=False, need_email=False, need_shipping_address=False, is_flexible=True):
		payload = {"chat_id": chat_id, "title": title, "description": description, "provider_token": provider_token, "prices": prices}
		if photo_url:
			payload["photo_url"] = photo_url
		payload["need_name"] = need_name
		payload["need_phone_number"] = need_phone_number
		payload["need_email"] = need_email
		payload["need_shipping_address"] = need_shipping_address
		payload["is_flexible"] = is_flexible

		return self.request(Route("POST", "sendInvoice", self.token), json=payload)

	def edit_message(self, chat_id, message_id, text, *, components=None):
		payload = {
			"chat_id": chat_id,
			"message_id": message_id,
			"text": text
		}
		if components:
			payload["reply_markup"] = components

		return self.request(Route("POST", "editMessageText", self.token), json=payload)

	def delete_message(self, chat_id, message_id):
		payload = {
			"chat_id": chat_id,
			"message_id": message_id
		}
		return self.request(Route("GET", "deletemessage", self.token), params=payload)

	def get_updates(self, offset=None, limit=None):
		payload = {}
		if offset:
			payload["offset"] = offset
		if limit:
			payload["limit"] = limit
		return self.request(Route("POST", "getupdates", self.token), json=payload)

	def delete_webhook(self):
		return self.request(Route("GET", "deleteWebhook", self.token))

	def get_bot(self):
		return self.request(Route("GET", "getme", self.token))

	def get_chat(self, chat_id):
		return self.request(Route("GET", "getchat", self.token), params=dict(chat_id=chat_id))

	def leave_chat(self, chat_id):
		return self.request(Route("GET", "leaveChat", self.token), params=dict(chat_id=chat_id))

	def get_chat_administrators(self, chat_id):
		return self.request(Route("GET", "getChatAdministrators", self.token), params=dict(chat_id=chat_id))

	def get_chat_members_count(self, chat_id):
		return self.request(Route("GET", "getChatAdministrators", self.token), params=dict(chat_id=chat_id))

	def get_chat_member(self, chat_id, member_id):
		return self.request(Route("GET", "getChatMember", self.token), params=dict(chat_id=chat_id, user_id=member_id))

	def invite_to_chat(self, chat_id, user_id):
		return self.request(Route("GET", "InviteUser", self.token), json=dict(chat_id=chat_id, user_id=user_id))



"""
    MIT License

    Copyright (c) 2022 kian Ahmadian

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from bale import Bot, Message, User, Photo


__all__ = (
    "Chat",
    "ChatType"
)

class ChatType:
    """This object indicates a Chat Type.

    .. container:: operations
        .. describe:: x == y
            Checks if two chat type are equal.
        .. describe:: x != y
            Checks if two chat type are not equal.
    """
    PRIVATE = "private"
    GROUP = "group"

    __slots__ = (
        "_type",
    )

    def __init__(self, _type: str):
        self._type = _type

    def is_private_chat(self):
        """bool:
            Return ``True`` if Chat Type is Private"""
        return self._type == self.PRIVATE

    def is_group_chat(self):
        """bool:
            Return ``True`` if Chat Type is Group"""
        return self._type == self.GROUP

    def __eq__(self, other):
        return self._type == other

    def __ne__(self, other):
        return not self.__eq__(other)

class Chat:
    """This object indicates a chat.

    Attributes
    ----------
        chat_id: str
            Chat ID.
        type: :class:`bale.ChatType`
            Chat Type.
        title: str
            Chat Title.
        username: str
            Chat Username (for DM or PV).
        first_name: str
            First name Chat (for DM or PV).
        last_name: str
            Last name Chat (for DM or PV).
        pinned_message: :class:`bale.Message`
            Pinned messages in chat. Defaults to None.
        all_members_are_administrators: bool
            Does everyone have admin access?. Defaults to True. (for Group)
        bot: :class:`bale.Bot`
            Bot Object. Defaults to None.

    .. container:: operations
        .. describe:: x == y
            Checks if two chat are equal.
        .. describe:: x != y
            Checks if two chat are not equal.
    """
    __slots__ = (
        "chat_id",
        "_type",
        "title",
        "username",
        "first_name",
        "last_name",
        "pinned_message",
        "all_members_are_administrators",
        "bot"
    )

    def __init__(self, chat_id: int | str, _type: "ChatType", title: str, username: str, first_name: str, last_name: str,
                 pinned_message: Message | None = None, all_members_are_administrators: bool = True, bot: 'Bot' = None):
        self.chat_id = chat_id
        self._type = _type
        self.title = title
        self.username = username
        self.first_name = first_name
        self.last_name = last_name
        self.pinned_message = pinned_message
        self.all_members_are_administrators = all_members_are_administrators
        self.bot = bot

    @property
    def type(self):
        """Get chat type"""
        return self._type

    @property
    def mention(self) -> str | None:
        """Optional[:class:`str`]"""
        return ("@" + self.username) if self.username else None

    async def send(self, text: str, components=None):
        """
        For the documentation of the arguments, please see :meth:`bale.Bot.send_message`.

        Parameters
        -----------
            text: str
                Message content
            components: Optional[:class:`bale.Components` | :class:`bale.RemoveComponents`]
                Message components
        Returns
        --------
            :class:`bale.Message`
        """
        return await self.bot.send_message(self, text, components=components)

    async def send_photo(self, photo: bytes | str | "Photo", caption: str = None):
        """
        For the documentation of the arguments, please see :meth:`bale.Bot.send_photo`.

        Parameters
        ----------
            photo: :class:`bytes` | :class:`str` | :class:`bale.Photo`
                Photo
            caption: str
                Message caption.
        Raises
        ------
            :class:`bale.Error`
        Returns
        -------
            Optional[:class:`bale.Message`]
                On success, the sent Message is returned.
        """
        return await self.bot.send_photo(self, photo, caption)

    async def leave(self):
        """
        For the documentation of the arguments, please see :meth:`bale.Bot.leave_chat`.

        Returns:
            :bool:
                On success, ``True``.
        """
        return await self.bot.leave_chat(self)

    async def add_user(self, user: "User"):
        """
        For the documentation of the arguments, please see :meth:`bale.Bot.invite_to_chat`.

        Parameters
        ----------
            user: :class:`bale.User`
                user
        Returns
        -------
            bool
                One success, ``True``.
        """
        return await self.bot.invite_to_chat(self, user)

    async def get_chat_member(self, user: "User"):
        """
        For the documentation of the arguments, please see :meth:`bale.Bot.get_chat_member`.

        Parameters
        ----------
            user: :class:`bale.User`
                User
        Returns
        -------
            Optional[:class:`bale.ChatMember`]
                On success, The chat member is retuened.
        """
        return await self.bot.get_chat_member(self, user)

    async def get_chat_members_count(self):
        """
        For the documentation of the arguments, please see :meth:`bale.Bot.get_chat_members_count`.

        Returns
        -------
            Optional[:class:`int`]
                On success, The count of chat members is returned.
        """
        return await self.bot.get_chat_members_count(self)

    async def get_chat_administrators(self):
        """
        For the documentation of the arguments, please see :meth:`bale.Bot.get_chat_administrators`.

        Raises
        ------
            :class:`bale.Error`
        Returns
        -------
            Optional[List[:class:`bale.ChatMember`]]
                On success, The chat members is returned.
        """
        return await self.bot.get_chat_administrators(self)

    @classmethod
    def from_dict(cls, data: dict, bot):
        return cls(bot=bot, chat_id=data.get("id"), _type=ChatType(data.get("type")), title=data.get("title"),
                   username=data.get("username"), first_name=data.get("first_name"), last_name=data.get("last_name"),
                   pinned_message=Message.from_dict(bot=bot, data=data.get("pinned_message")) if data.get("pinned_message") else None,
                   all_members_are_administrators=data.get("all_members_are_administrators", True))

    def to_dict(self):
        data = {
            "id": self.chat_id,
            "type": self.type,
            "title": self.title,
            "username": self.username,
            "first_name": self.first_name,
            "last_name": self.last_name
        }
        if self.pinned_message:
            data["pinned_message"] = self.pinned_message.to_dict()

        return data

    def __str__(self):
        return (str(self.username) + "#" + str(self.chat_id) if self.username else str(self.first_name) + " " + str(
            self.last_name))

    def __eq__(self, other):
        return isinstance(other, Chat) and self.chat_id == other.chat_id

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__str__())

    def __repr__(self):
        return f"<Chat first_name={self.first_name} last_name={self.last_name} user_id={self.chat_id} username={self.username} title={self.title}>"



from aiohttp import ClientResponse
from typing import Any
from json import loads
from json.decoder import JSONDecodeError


async def json_or_text(response: "ClientResponse"):
	text = await response.text()

	try:
		json = loads(text)
	except JSONDecodeError:
		return text
	else:
		return json

class ResponseParser:
	"""a Parser for parse http response.

	Attributes
	==========
	ok: bool
		Status of the response

	result: Any
		Result of the response

	error_code: Optional[:class:`int`]
		Error Code.
		``None`` when no error

	description: Optional[:class:`str`]
		Description of the error.
		``None`` when no error.

	_raw: dict
		Raw of the response data
	"""

	__slots__ = (
		"result",
		"error_code",
		"description",
		"ok",
		"_raw"
	)

	def __init__(self, ok: bool, result: Any = None, error_code: int = None, description: str = None, raw: dict = None):
		self.ok = ok

		self.result = result
		self.error_code = error_code
		self.description = description
		self._raw = raw

	@classmethod
	async def from_response(cls, data: "ClientResponse"):
		if data.status == 404:
			return cls(False, raw=dict())

		data = await json_or_text(data)

		if type(data) == str:
			return cls(False, description=data, raw=dict(description=data))
		else:
			return cls(data.get("ok", False), data.get("result"), data.get("error_code"), data.get("description"), data)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from bale import Bot


class Document:
	"""This object shows a Document.

    Attributes
    ----------
        file_id: Optional[:class:`str`]
        file_name: Optional[:class:`str`]
        mime_type: Optional[:class:`str`]
        file_size: Optional[:class:`int`]
    """
	__slots__ = (
		"file_id",
		"file_name",
		"mime_type",
		"file_size",
		"bot"
	)

	def __init__(self, file_id: str = None, file_name: str = None, mime_type: str = None, file_size: int = None,
	             bot: "Bot" = None):
		self.file_id = file_id if file_id is not None else None
		self.file_name = file_name if file_name is not None else None
		self.mime_type = mime_type if mime_type is not None else None
		self.file_size = file_size if file_size is not None else None
		self.bot = bot

	async def read(self):
		"""Read the Document.

        Raises
        ------
            :class:`bale.Error`
        Returns
        -------
            :class:`bytes`

        """
		return await self.bot.http.get_file(self.file_id)

	async def save(self, file_name):
		"""Save the Document.

        Parameters
        ----------
            file_name: str

        Raises
        ------
            :class:`bale.Error`

        """
		data = await self.read()
		with open(file_name, 'wb') as file:
			return file.write(data)

	@classmethod
	def from_dict(cls, data: dict, bot: "Bot" = None):
		return cls(file_id=data.get("file_id"), file_name=data.get("file_name"),
		           mime_type=data.get("mime_type"), file_size=data.get("file_size"), bot=bot)

	def to_dict(self):
		data = { "file_id": self.file_id, "file_name": self.file_name, "mime_type": self.mime_type,
		         "file_size": self.file_size }

		return data



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from bale import Bot
from bale import User, Message

__all__ = (
    "CallbackQuery"
)


class CallbackQuery:
    """This object represents an incoming callback query from a callback button in an inline keyboard.

    Parameters
    -----------
        callback_id: int
            Callback Query ID
        data: str
            Callback Data
        message: :class:`bale.Message`
            Callback Message
        inline_message_id: str
            Callback inline message id
        from_user: :class:`bale.User`
            Callback The user who gave the callback.
        bot: Optional[:class:`bale.Bot`]
            Bot object.
    """
    __slots__ = (
        "message",
        "inline_message_id",
        "from_user",
        "data",
        "bot",
        "callback_id"
    )

    def __init__(self, callback_id: int = None, data: str = None, message: "Message" = None, inline_message_id: str = None,
                 from_user: "User" = None, bot: "Bot" = None):
        self.data = data
        self.callback_id = callback_id
        self.message = message
        self.inline_message_id = inline_message_id
        self.from_user = from_user
        self.bot = bot

    @property
    def user(self):
        """Aliases for :attr:`from_user`"""
        return self.from_user

    @classmethod
    def from_dict(cls, data: dict, bot: "Bot"):
        return cls(bot=bot, data=data.get("data"), callback_id=data.get("id"), message=Message.from_dict(data.get("message"), bot=bot),
                   inline_message_id=data.get("inline_message_id"), from_user=User.from_dict(bot=bot, data=data.get("from")))

    def to_dict(self):
        data = {
            "id": self.callback_id
        }

        if self.data:
            data["data"] = self.data
        if self.inline_message_id:
            data["inline_message_id"] = self.inline_message_id
        if self.message:
            data["message"] = self.message.to_dict()
        if self.from_user:
            data["from_user"] = self.from_user.to_dict()

        return data

    def __repr__(self):
        return f"<CallbackQuery inline_message_id={self.inline_message_id} message={self.message} from_user={self.from_user} data={self.data}>"
